package com.apifon.sdk;

import com.apifon.sdk.exception.ApifonRestException;
import com.apifon.sdk.exception.OAuthException;
import com.apifon.sdk.exception.OAuthTokenException;
import com.apifon.sdk.model.request.OAuthConfigurationBuilder.OAuthConfiguration;
import com.apifon.sdk.resource.AbstractResource;
import com.apifon.sdk.resource.helpers.OAuthCredentials;
import com.apifon.sdk.response.ApiErrorResponse;
import com.apifon.sdk.security.SignedRequests;
import com.google.gson.Gson;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.TimeZone;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpDelete;
import org.apache.http.client.methods.HttpEntityEnclosingRequestBase;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpPut;
import org.apache.http.client.methods.HttpRequestBase;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.util.EntityUtils;
//import org.apache.http.ssl.SSLContextBuilder;

/**
 * This class is used to make HTTP requests to APIFON Rest API through java. To
 * send API requests first call the static method setCredentials(token,
 * sectetKey) using your unique API token and your secret key that you have
 * gained from the Mookee application and then you can use the available
 * Resources in order for you to send your requests.
 */
public class Mookee {

    private static Boolean staging = false;
    private static volatile Mookee instance = null;

    private static String base_url = null;

    private static final String PRODUCTION_URL = "https://ars.apifon.com";
    private static volatile String STAGING_URL = "http://ars-staging.apifon.com";
    private static final String IDENTITY_PRODUCTION_SERVER_URL = "https://ids.apifon.com/";
    private static volatile String IDENTITY_STAGING_SERVER_URL = "https://ids-staging.apifon.com/";

    private static Map<String, ArrayList> credentials = new HashMap();
    private static String key = "";
    private static String activeOAuthIdentifier = null;
    private static boolean autoRefreshTokens = false;
    private static SignedRequests signer = null;

    private Mookee() {
    }

    public static Mookee getInstance() {
        // Check if instance already exists
        if (instance == null) {
            instance = new Mookee();
        }
        return instance;
    }

    /**
     * Sets the currently active OAuth client identifier.<br><br>
     * <b>Note</b>: Setting OAuth mode will force Mookee to dispatch <b>ALL</b>
     * requests using the given OAuth client credentials for this identifier.
     *
     * @param identifier The identifier to set active as OAuth
     * @param autoRefreshTokens If set to true Mookee will try to transparently
     * refresh tokens if it detects that the token has expired/revoked.
     * Otherwise Mookee will try to use active client credentials and throw an
     * exception if a token has expired.
     * @throws IllegalArgumentException if the given identifier is unknown or is
     * not OAuth enabled.
     * @see #setMookeeToHMACTokenMode() Turn Mookee into HMAC token/secret mode.
     * @see #addOAuthResource(java.lang.String,
     * com.apifon.sdk.model.request.OAuthConfigurationBuilder.OAuthConfiguration)
     * Add OAuth configuration for an identifier.
     */
    public static void setMookeeToOAuthMode(String identifier, boolean autoRefreshTokens) throws IllegalArgumentException {
        if (isOAuthEnabled(identifier)) {
            Mookee.activeOAuthIdentifier = identifier;
            Mookee.autoRefreshTokens = autoRefreshTokens;
            return;
        }
        throw new IllegalArgumentException("Identifier " + identifier + " is not a valid OAuth identifier.");
    }

    /**
     * Sets mookee to the HMAC token/secret key mode.<br><br>
     * <b>Note</b>: Setting mookee on HMAC token mode will force Mookee to use
     * Token credentials for <b>ALL</b> requests.<br>
     * This will <b>NOT</b> invalidate any active OAuth2 token credentials already registered.
     *
     * @see #setMookeeToOAuthMode(java.lang.String, boolean) Enable OAuth mode in Mookee
     * @see #revokeTokens(java.lang.String) Revoke existing OAuth tokens
     * @see #removeOAuthResource(java.lang.String, boolean) Remove OAuth resource
     */
    public static void setMookeeToHMACTokenMode() {
        activeOAuthIdentifier = null;
    }

    /**
     * Set to True if you want to use our Staging Server.
     *
     * @param enable A boolean value.
     * @param host A testing host, useful for development. May be null in which case 
     * Apifon's Staging Server URL is used.
     * @param identityServerHost A host for the identity server, useful for
     * development. May be null in which case Apifon's Identity Server URL is used.
     */
    public static void setStaging(Boolean enable, String host, String identityServerHost) {
        staging = enable;
        if (host != null) {
            STAGING_URL = host;
        }
        if (identityServerHost != null) {
            IDENTITY_STAGING_SERVER_URL = identityServerHost;
        }
    }

    /**
     * Set to True if you want to use our Staging Server.
     *
     * @param enable A boolean value.
     * @param host A testing host, useful for development. May be null.
     */
    public static void setStaging(Boolean enable, String host) {
        Mookee.setStaging(enable, host, null);
    }

    /**
     * Adds a given OAuth configuration for the given identifier.<br>
     * This will overwrite any existing configuration AND revoke any (if any)
     * existing tokens for this identifier.
     *
     * @param identifier
     * @param oauthConfiguration
     * @throws com.apifon.sdk.exception.OAuthException
     */
    public static void addOAuthResource(String identifier, OAuthConfiguration oauthConfiguration) throws OAuthException {
        OAuthRegistry.INSTANCE.addOAuthResource(identifier, oauthConfiguration);
    }

    /**
     * Removes the OAuthResource with the given identifier. This will also
     * revoke any active tokens and refresh tokens (if any) for the given
     * identifier.
     *
     * @param identifier The identifier to remove
     * @param revokeTokens true if you want to revoke any access and refresh
     * tokens before removal, false otherwise.
     * @throws OAuthException if a token revocation fails.
     */
    public static void removeOAuthResource(String identifier, boolean revokeTokens) throws OAuthException {
        OAuthRegistry.INSTANCE.removeOAuthResource(identifier, revokeTokens);
    }

    /**
     * Tries to revoke tokens (access token and refresh token if available) for
     * the given identifier.<br><br>
     * Note that after this method completes successfully your tokens will be
     * invalid and any authorization must be performed again.<br>
     * This will NOT remove the OAuthResource from the Registry, you can call
     * {@link #removeOAuthResource(java.lang.String, boolean)} for that.
     *
     * @param identifier the identifier to look for
     * @throws OAuthException in case an exception occurs during token
     * revocation.
     */
    public static void revokeTokens(String identifier) throws OAuthException {
        OAuthRegistry.INSTANCE.revokeTokens(identifier);
    }

    /**
     * @param identifier The identifier to look for.
     * @return true if OAuth is enabled for the given identifier, false
     * otherwise.
     */
    public static boolean isOAuthEnabled(String identifier) {
        return OAuthRegistry.INSTANCE.isOAuthEnabled(identifier);
    }

    /**
     * Tries to retrieve access token for the client credentials OAuth grant.
     *
     * @param identifier
     * @throws OAuthException If the resource is not OAuth enabled or an
     * exception occurs.
     */
    public static void retrieveClientCredentialsToken(String identifier) throws OAuthException {
        OAuthRegistry.INSTANCE.retrieveTokensForClientCredential(identifier);
    }

    /**
     * Tries to retrieve access token for the client credentials OAuth grant and
     * returns a copy of the tokens.
     *
     * @param identifier
     * @return A copy of the OAuthCredentials generated.
     * @throws OAuthException If the resource is not OAuth enabled or an
     * exception occurs.
     */
    public static OAuthCredentials retrieveClientCredentialsAndGetToken(String identifier) throws OAuthException {
        return OAuthRegistry.INSTANCE.retrieveTokensForClientCredentialAndGet(identifier);
    }

    /**
     * Refresh active access token. This will revoke any active tokens and issue
     * a new one.
     *
     * @param identifier
     * @throws OAuthException If the resource is not OAuth enabled or an
     * exception occurs.
     */
    public static void refreshAccessToken(String identifier) throws OAuthException {
        OAuthRegistry.INSTANCE.refreshTokens(identifier);
    }

    public static String dispatch(AbstractResource resource) {
        if (staging) {
            base_url = STAGING_URL;
        } else {
            base_url = PRODUCTION_URL;
        }
        String endpoint = resource.getCurEndpoint();
        String body = resource.getCurBody();
        String method = resource.getCurMethod();
        HttpRequestBase request = null;
        switch (method) {
            case "GET":
                request = new HttpGet();
                break;
            case "POST":
                request = new HttpPost();
                break;
            case "PUT":
                request = new HttpPut();
                break;
            case "DELETE":
                request = new HttpDelete();
                break;
        }
        try {
            return Mookee.makeCall(endpoint, method, body, request);
        } catch (SecurityException ex) {
            Logger.getLogger(Mookee.class.getName()).log(Level.SEVERE, "401 Unauthorized");
        } catch (Exception ex) {
            Logger.getLogger(Mookee.class.getName()).log(Level.SEVERE, null, ex);
        }
        return "";
    }

    public static HttpResponse dispatchToResponse(AbstractResource resource) throws ApifonRestException {
        if (staging) {
            base_url = STAGING_URL;
        } else {
            base_url = PRODUCTION_URL;
        }
        String endpoint = resource.getCurEndpoint();
        String body = resource.getCurBody();
        String method = resource.getCurMethod();
        HttpRequestBase request = null;
        switch (method) {
            case "GET":
                request = new HttpGet();
                break;
            case "POST":
                request = new HttpPost();
                break;
            case "PUT":
                request = new HttpPut();
                break;
            case "DELETE":
                request = new HttpDelete();
                break;
            default:
                throw new ApifonRestException("Failed to understand method parameter " + method);
        }
        try {
            resource.addHeaders(request);

            HttpResponse resp;

            if (Mookee.activeOAuthIdentifier != null) {
                resp = Mookee.makeOauthCallToResponse(endpoint, body, request);
            } else {
                resp = Mookee.makeCallToResponse(endpoint, method, body, request);
            }

            if (resp.getStatusLine().getStatusCode() >= 400) {
                String jret = EntityUtils.toString(resp.getEntity());
                throw new ApifonRestException(new Gson().fromJson(jret, ApiErrorResponse.class));
            } else {
                return resp;
            }
        } catch (ApifonRestException ex) {
            throw ex;
        } catch (Exception ex) {
            throw new ApifonRestException(new IOException(ex));
        }
    }

    private static String makeCall(String endpoint, String method, String body, HttpRequestBase request) throws SecurityException, Exception {
        HttpResponse response;
        if (Mookee.activeOAuthIdentifier != null) {
            response = makeOauthCallToResponse(endpoint, body, request);
        } else {
            response = makeCallToResponse(endpoint, method, body, request);
        }
        HttpEntity httpEntity = response.getEntity();
        String jret = EntityUtils.toString(httpEntity);
        if (response.getStatusLine().getStatusCode() == 401) {
            throw new SecurityException("401 Unauthorized");
        }
        return jret;
    }

    /**
     * Adds your Credentials.
     *
     * @param key an association key.
     * @param tkn your ApiToken.
     * @param prKey your PrivateKey.
     */
    public static void addCredentials(String key, String tkn, String prKey) {
        ArrayList<String> list = new ArrayList();
        list.add(tkn);
        list.add(prKey);
        credentials.put(key, list);
    }

    /**
     * Chooses your Credential.
     *
     * @param k an association key.
     */
    public static void setActiveCredential(String k) {
        key = k;
    }

    private static HttpResponse makeCallToResponse(String endpoint, String method, String body, HttpRequestBase request) throws Exception {
        if (body == null) {
            body = "";
        } else {
            StringEntity input = new StringEntity(body, "UTF-8");
            input.setContentType("application/json");
            if (!body.equals("") && !body.isEmpty()) {
                ((HttpEntityEnclosingRequestBase) request).setEntity(input);
            }
        }
        String uri;
        if (endpoint.startsWith(base_url)) {
            uri = endpoint;
        } else {
            uri = base_url + endpoint;
        }
        request.setURI(new URI(uri));

        signer = new SignedRequests((String) credentials.get(key).get(1));
        SimpleDateFormat dateFormat = new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss z", Locale.US);
        dateFormat.setTimeZone(TimeZone.getTimeZone("GMT"));
        String strDate = dateFormat.format(new Date());
        HttpClient httpClient = HttpClientBuilder.create().build();

        URI tmUri = null;
        try {
            tmUri = new URI(endpoint);
        } catch (URISyntaxException ex) {
            Logger.getLogger(Mookee.class.getName()).log(Level.SEVERE, null, ex);
        }
        String toSign = method + "\n"
                + tmUri.getPath() + "\n"
                + body + "\n"
                + strDate;
        String signature = signer.hmac(toSign);
        request.setHeader("Content-Type", "application/json");
        request.setHeader("Authorization", "ApifonWS " + (String) credentials.get(key).get(0) + ":" + signature);
        request.setHeader("X-ApifonWS-Date", strDate);

        return httpClient.execute(request);
    }

    private static HttpResponse makeOauthCallToResponse(String endpoint, String body, HttpRequestBase request) throws Exception {
        if (body == null) {
            body = "";
        } else {
            StringEntity input = new StringEntity(body, "UTF-8");
            input.setContentType("application/json");

            if (!body.equals("") && !body.isEmpty()) {
                ((HttpEntityEnclosingRequestBase) request).setEntity(input);
            }
        }
        String uri;
        if (endpoint.startsWith(base_url)) {
            uri = endpoint;
        } else {
            uri = base_url + endpoint;
        }

        request.setURI(new URI(uri));
        HttpClient httpClient = HttpClientBuilder.create().build();
        request.setHeader("Authorization", "Bearer " + getActiveOAuthCredential());
        request.setHeader("Content-Type", "application/json");
        HttpResponse response = httpClient.execute(request);
        // Unauthorized to access API, check auto refresh and try to refresh token manually from IDS.
        if (response.getStatusLine().getStatusCode() == 401) {
            if (Mookee.autoRefreshTokens) {
                // Try to refresh tokens and resubmit request!
                try {
                    OAuthRegistry.INSTANCE.refreshTokens(Mookee.activeOAuthIdentifier);
                    request.setHeader("Authorization", "Bearer " + getActiveOAuthCredential());
                    HttpResponse response1 = httpClient.execute(request);
                    if (response1.getStatusLine().getStatusCode() == 401){
                        throw new OAuthTokenException("Status: 401" + System.lineSeparator() 
                                + "Error: " + response1.getFirstHeader("WWW-Authenticate").getValue());
                    }
                    return response1;
                } catch (OAuthException ex) {
                    throw ex;
                }
            } else {
                throw new OAuthException("Unauthorized to access API");
            }
        }
        return response;
    }

    private static String getActiveOAuthCredential() throws OAuthException {
        OAuthCredentials oauthCredentials = OAuthRegistry.INSTANCE.retrieveActiveTokens(Mookee.activeOAuthIdentifier);
        if (oauthCredentials.hasExpired() && Mookee.autoRefreshTokens) {
            OAuthRegistry.INSTANCE.refreshTokens(Mookee.activeOAuthIdentifier);
            return OAuthRegistry.INSTANCE.retrieveActiveTokens(Mookee.activeOAuthIdentifier).getAccessToken();
        }
        return oauthCredentials.getAccessToken();
    }

    public String getIdentityServerUrl() {
        if (Mookee.staging != null && Mookee.staging) {
            return Mookee.IDENTITY_STAGING_SERVER_URL;
        }
        return Mookee.IDENTITY_PRODUCTION_SERVER_URL;
    }

}
