/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package com.apifon.sdk;

import com.apifon.sdk.exception.OAuthException;
import com.apifon.sdk.model.request.OAuthConfigurationBuilder;
import com.apifon.sdk.resource.OAuthResource;
import com.apifon.sdk.resource.helpers.OAuthCredentials;
import java.io.IOException;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Handles OAuth credentials
 *
 * @author Apifon
 */
public enum OAuthRegistry {
    INSTANCE;

    private final Map<String, OAuthResource> oauthRegistry;

    private OAuthRegistry() {
        oauthRegistry = new ConcurrentHashMap<>();
    }

    /**
     * Tries to add an OAuthResource with the given identifier.<br><br>
     * <b>This will revoke and remove any existing OAuthResource and
     * OAuthCredentials for this identifier.</b>
     *
     * @param identifier
     * @param oauthConfiguration
     */
    public void addOAuthResource(String identifier, OAuthConfigurationBuilder.OAuthConfiguration oauthConfiguration) throws OAuthException {
        OAuthResource oAuthResource = new OAuthResource(oauthConfiguration);
        if (oauthRegistry.containsKey(identifier)) {
            removeOAuthResource(identifier, true);
        }
        oauthRegistry.put(identifier, oAuthResource);
    }

    /**
     * Removes the given OAuth info related to the given identifier from the
     * registry and revokes any tokens (if any)<br><br>
     * No action will be performed if the identifier is invalid.
     *
     * @param identifier the identifier of the OAuth information to remove
     * @param revokeTokens true if you want to revoke any existing tokens.
     * @throws com.apifon.sdk.exception.OAuthException in case that token
     * revocations fails
     */
    public void removeOAuthResource(String identifier, boolean revokeTokens) throws OAuthException {
        if (oauthRegistry.containsKey(identifier)) {
            if (revokeTokens) {
                revokeTokens(identifier);
            }
            oauthRegistry.remove(identifier);
        }
    }

    /**
     * @param identifier the identifier to look for
     * @return true if the given identifier is OAuth enabled, false otherwise.
     */
    public boolean isOAuthEnabled(String identifier) {
        return oauthRegistry.containsKey(identifier);
    }

    /**
     * Tries to revoke tokens (access token and refresh token if available) for
     * the given resource.<br><br>
     * Note that after this method completes successfully your tokens will be
     * invalid and any authorization must be performed again.
     *
     * @param identifier the identifier to look for
     * @throws OAuthException in case an exception occurs during token
     * revocation.
     */
    public void revokeTokens(String identifier) throws OAuthException {
        validateOAuthEnabled(identifier);
        oauthRegistry.get(identifier).revokeTokens();
    }

    public OAuthResource getResourceForIdentifier(String identifier) throws OAuthException {
        validateOAuthEnabled(identifier);
        return oauthRegistry.get(identifier);
    }

    public void retrieveTokensForClientCredential(String identifier) throws OAuthException {
        validateOAuthEnabled(identifier);
        OAuthResource resource = oauthRegistry.get(identifier);
        resource.getTokensForClientCredentialsGrant();
    }

    public OAuthCredentials retrieveTokensForClientCredentialAndGet(String identifier) throws OAuthException {
        validateOAuthEnabled(identifier);
        OAuthResource resource = oauthRegistry.get(identifier);
        resource.getTokensForClientCredentialsGrant();
        try {
            return resource.getOauthCredentials().clone();
        } catch (CloneNotSupportedException ex) {
            throw new OAuthException(new IOException(ex));
        }
    }

    public void refreshTokens(String identifier) throws OAuthException {
        validateOAuthEnabled(identifier);
        oauthRegistry.get(identifier).refreshToken();
    }

    public OAuthCredentials retrieveActiveTokens(String identifier) throws OAuthException {
        if (isOAuthEnabled(identifier)) {
            try {
                return getResourceForIdentifier(identifier).getOauthCredentials().clone();
            } catch (CloneNotSupportedException ex) {
                throw new OAuthException(new IOException(ex));
            }
        }
        throw new OAuthException("Active tokens for " + identifier + " not located.");
    }

    private void validateOAuthEnabled(String identifier) throws OAuthException {
        if (!isOAuthEnabled(identifier)) {
            throw new OAuthException("Identifier " + identifier + " doesn't have OAuth enabled.");
        }
    }

}
