package com.apifon.sdk.model.request;

import com.apifon.sdk.resource.helpers.GroupOptionsOrientation;
import com.apifon.sdk.resource.helpers.GroupType;
import com.apifon.sdk.resource.helpers.IJsonSerializable;
import com.apifon.sdk.resource.helpers.SelectionType;
import com.apifon.sdk.response.GroupResponse;
import com.google.gson.Gson;

import java.util.Objects;

public class GroupRequestBuilder {

    private final transient String listId;
    private final transient String groupField;
    private String name;
    private GroupType type;
    private SelectionType selectionType;
    private GroupOptionsOrientation groupOptionsOrientation;
    private boolean internal;

    private GroupRequestBuilder(String listId, GroupType type, String name, String groupField){
        this.listId = listId;
        this.type = type;
        this.name = name;
        this.groupField = groupField;
    }

    /**
     * Constructs a GroupRequestBuilder for add requests with the required
     * fields.<br>
     *
     * @param listId the list id. May NOT be null.
     * @param type the group type. May NOT be null.
     * @param name the group name. May NOT be null.
     * @return
     * @throws NullPointerException if any parameter is null.
     */
    public static GroupRequestBuilder getAddGroupRequestBuilder(String listId, GroupType type, String name){
        Objects.requireNonNull(listId,"list id is empty");
        Objects.requireNonNull(type,"type is empty");
        Objects.requireNonNull(name,"name is empty");

        return new GroupRequestBuilder(listId,type,name,null);
    }

    /**
     * Constructs a GroupRequestBuilder for update requests with the id
     * given.<br>
     *
     * @param listId the list id. May NOT be null.
     * @param type the group type. May NOT be null.
     * @param name the group name. May NOT be null.
     * @param groupField the group field. May NOT be null.
     * @return
     * @throws NullPointerException if any parameter is null.
     */
    public static GroupRequestBuilder getUpdateGroupRequestBuilder(String listId,String groupField, GroupType type, String name){
        Objects.requireNonNull(listId,"list id is empty");
        Objects.requireNonNull(type,"type is empty");
        Objects.requireNonNull(name,"name is empty");
        Objects.requireNonNull(groupField,"groupField is empty");

        return new GroupRequestBuilder(listId,type,name,groupField);
    }

    /**
     * Constructs a GroupRequestBuilder with fields set directly from the API
     * response.<br>
     *
     * @param groupResponse
     * @return
     */
    public static GroupRequestBuilder getUpdateGroupRequestBuilder(GroupResponse groupResponse, String listId){
        Objects.requireNonNull(listId,"list id is empty");
        Objects.requireNonNull(groupResponse,"groupResponse is empty");
        Objects.requireNonNull(groupResponse.getField(),"field is empty");
        return new GroupRequestBuilder(listId,groupResponse.getType(),groupResponse.getName(),groupResponse.getField());
    }

    /**
     * Sets the groups name. May NOT be null.
     *
     * @param name
     * @return this for chaining.
     * @throws NullPointerException if name is null.
     */
    public GroupRequestBuilder withName(String name){
        Objects.requireNonNull(name,"name is empty");
        this.name = name;
        return this;
    }

    /**
     * Sets the groups type. May NOT be null.
     *
     * @param type
     * @return this for chaining.
     * @throws NullPointerException if type is null.
     */
    public GroupRequestBuilder withType(GroupType type){
        Objects.requireNonNull(type,"type is empty");
        this.type = type;
        return this;
    }

    /**
     * Sets the groups internal. May NOT be null.
     *
     * @param internal
     * @return this for chaining.
     * @throws NullPointerException if internal is null.
     */
    public GroupRequestBuilder withInternal(boolean internal){
        Objects.requireNonNull(internal,"internal is empty");
        this.internal = internal;
        return this;
    }

    /**
     * Sets the groups selectionType. May NOT be null.
     *
     * @param selectionType
     * @return this for chaining.
     * @throws NullPointerException if selectionType is null.
     */
    public GroupRequestBuilder withSelectionType(SelectionType selectionType){
        Objects.requireNonNull(selectionType,"selectionType is empty");
        this.selectionType = selectionType;
        return this;
    }

    /**
     * Sets the groups groupOptionsOrientation. May NOT be null.
     *
     * @param groupOptionsOrientation
     * @return this for chaining.
     * @throws NullPointerException if groupOptionsOrientation is null.
     */
    public GroupRequestBuilder withGroupOptionsOrientation(GroupOptionsOrientation groupOptionsOrientation){
        Objects.requireNonNull(groupOptionsOrientation,"groupOptionsOrientation is empty");
        this.groupOptionsOrientation = groupOptionsOrientation;
        return this;
    }

    public GroupRequest build(){
        return new GroupRequest(this);
    }

    public static final class GroupRequest implements IJsonSerializable {
        private final transient String listId;
        private final transient String groupField;
        private final String name;
        private final GroupType type;
        private final SelectionType selectionType;
        private final GroupOptionsOrientation groupOptionsOrientation;
        private final boolean internal;

        public GroupRequest(GroupRequestBuilder builder){
            this.listId = builder.listId;
            this.groupField = builder.groupField;
            this.name = builder.name;
            this.type = builder.type;
            this.internal = builder.internal;
            this.selectionType = builder.selectionType;
            this.groupOptionsOrientation = builder.groupOptionsOrientation;
        }

        public String getListId() {
            return listId;
        }

        public String getGroupField() {
            return groupField;
        }

        public String getName() {
            return name;
        }

        public GroupType getType() {
            return type;
        }

        public SelectionType getSelectionType() {
            return selectionType;
        }

        public GroupOptionsOrientation getGroupOptionsOrientation() {
            return groupOptionsOrientation;
        }

        public boolean isInternal() {
            return internal;
        }

        @Override
        public String getBody() {
            return new Gson().toJson(this);
        }
    }
}
