/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package com.apifon.sdk.model.request;

import com.apifon.sdk.resource.helpers.OAuthGrantType;
import com.apifon.sdk.resource.helpers.OpenIDScope;
import com.apifon.sdk.resource.helpers.S256PKCEChallenge;
import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.util.Base64;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Objects;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Collectors;

/**
 *
 * @author Apifon
 */
public class OAuthConfigurationBuilder {

    private final OAuthGrantType grantType;
    private final String clientId;
    private final Set<OpenIDScope> scope;
    private String clientSecret;
    private S256PKCEChallenge pkceChallenge;
    private String redirectUri;
    private String state;

    private OAuthConfigurationBuilder(String clientId, OAuthGrantType grantType) {
        this.clientId = clientId;
        this.grantType = grantType;
        this.scope = new HashSet<>();
    }

    public static OAuthConfigurationBuilder getClientCredentialsBuilder(String clientId, String clientSecret) {
        OAuthConfigurationBuilder oauthRequestBuilder = new OAuthConfigurationBuilder(clientId, OAuthGrantType.CLIENT_CREDENTIALS);
        oauthRequestBuilder.clientSecret = clientSecret;
        return oauthRequestBuilder;
    }

    public OAuthConfigurationBuilder withScope(OpenIDScope scope) {
        this.scope.add(scope);
        return this;
    }

    public OAuthConfigurationBuilder withScopes(Collection<OpenIDScope> scopes) {
        this.scope.addAll(scopes);
        return this;
    }

    public OAuthConfigurationBuilder withRedirectUri(String redirectUri) {
        this.redirectUri = redirectUri;
        return this;
    }

    public OAuthConfigurationBuilder withRedirectUri(URI redirectUri, boolean forceUrlEncoding) {
        this.redirectUri = redirectUri.toString();
        return this;
    }
    
    /**
     * Sets a state parameter for client side validation.<br>
     * Note: If you do not explicit set a state a random one is generated for you.
     * @param state
     * @return this for chaining
     * @throws NullPointerException if state is null.
     */
    public OAuthConfigurationBuilder withState(String state) throws NullPointerException {
        Objects.requireNonNull(state, "state parameter can not be null");
        this.state = state;
        return this;
    }

    public OAuthConfiguration build() {
        return new OAuthConfiguration(this);
    }

    public class OAuthConfiguration {

        private final OAuthGrantType grantType;
        private final String clientId;
        private final Set<OpenIDScope> scope;
        private final String clientSecret;
        private final S256PKCEChallenge pkceChallenge;
        private final String redirectUri;
        private final String state;

        public OAuthConfiguration(OAuthConfigurationBuilder builder) {
            this.grantType = builder.grantType;
            this.clientId = builder.clientId;
            this.scope = new HashSet<>(builder.scope);
            this.clientSecret = builder.clientSecret;
            if (builder.pkceChallenge != null) {
                this.pkceChallenge = new S256PKCEChallenge(builder.pkceChallenge.getCodeVerifier());
            } else {
                this.pkceChallenge = null;
            }
            this.redirectUri = builder.redirectUri;
            if (builder.state == null) {
                this.state = Base64.getUrlEncoder().encodeToString(UUID.randomUUID().toString().getBytes(StandardCharsets.UTF_8));
            } else {
                this.state = builder.state;
            }
        }

        public OAuthGrantType getGrantType() {
            return grantType;
        }

        public String getClientId() {
            return clientId;
        }

        public Set<OpenIDScope> getScope() {
            return Collections.unmodifiableSet(this.scope);
        }

        public String getScopesAsString() {
            return this.scope
                    .stream()
                    .map((k) -> k.getScopeValue())
                    .collect(Collectors.joining("+"));
        }

        public String getClientSecret() {
            return clientSecret;
        }

        public S256PKCEChallenge getPkceChallenge() {
            return pkceChallenge;
        }

        public String getRedirectUri() {
            return redirectUri;
        }

        public String getState() {
            return state;
        }
    }

}
