package com.apifon.sdk.model.request;

import com.apifon.sdk.resource.helpers.GroupOptionsOrientation;
import com.apifon.sdk.resource.helpers.GroupType;
import com.apifon.sdk.resource.helpers.IJsonSerializable;
import com.apifon.sdk.resource.helpers.SelectionType;
import com.apifon.sdk.response.GroupResponse;
import com.google.gson.Gson;
import com.google.gson.annotations.SerializedName;

import java.util.List;
import java.util.Objects;
import java.util.logging.Level;
import java.util.logging.Logger;

public class SubgroupRequestBuilder {

    private final transient String listId;
    private final transient String groupField;
    private final String name;
    private final List<String> names;
    private final String optionsField;

    private SubgroupRequestBuilder(String listId, String optionsField, String groupField, String name, List<String> names){
        this.listId = listId;
        this.optionsField = optionsField;
        this.name = name;
        this.names = names;
        this.groupField = groupField;
    }

    /**
     * Constructs a SubgroupRequestBuilder for add requests with the required
     * fields.<br>
     *
     * @param listId the list id. May NOT be null.
     * @param groupField the groupField. May NOT be null.
     * @param name the subgroup name. May NOT be null.
     * @return
     * @throws NullPointerException if any parameter is null.
     */
    public static SubgroupRequestBuilder getAddSubgroupsRequestBuilder(String listId, String groupField, String name){
        Objects.requireNonNull(listId,"list id is empty");
        Objects.requireNonNull(groupField,"groupField is empty");
        Objects.requireNonNull(name,"name is empty");

        return new SubgroupRequestBuilder(listId,null,groupField,name,null);
    }

    /**
     * Constructs a SubgroupRequestBuilder for add requests with the required
     * fields.<br>
     *
     * @param listId the list id. May NOT be null.
     * @param groupField the groupField. May NOT be null.
     * @param names the subgroup names. May NOT be null.
     * @return
     * @throws NullPointerException if any parameter is null.
     */
    public static SubgroupRequestBuilder getAddSubgroupsRequestBuilder(String listId, String groupField, List<String> names){
        Objects.requireNonNull(listId,"list id is empty");
        Objects.requireNonNull(groupField,"groupField is empty");
        Objects.requireNonNull(names,"name is empty");

        return new SubgroupRequestBuilder(listId,null,groupField,null,names);
    }
    /**
     * Constructs a SubgroupRequestBuilder for update requests with the id
     * given.<br>
     *
     * @param listId the list id. May NOT be null.
     * @param optionField the option field . May NOT be null.
     * @param name the group name. May NOT be null.
     * @param groupField the group field. May NOT be null.
     * @return
     * @throws NullPointerException if any parameter is null.
     */
    public static SubgroupRequestBuilder getUpdateSubgroupRequestBuilder(String listId,  String groupField, String optionField, String name){
        Objects.requireNonNull(listId,"list id is empty");
        Objects.requireNonNull(groupField,"groupField is empty");
        Objects.requireNonNull(name,"name is empty");
        Objects.requireNonNull(optionField,"optionField is empty");

        return new SubgroupRequestBuilder(listId,optionField,groupField,name,null);
    }

    public SubgroupRequestBuilder.SubgroupRequest build(){
        if(name == null){
            return new SubgroupRequestMultiple(this);
        }
        else if(names == null){
            return new SubgroupRequestSingle(this);
        }
        else {
            Logger.getLogger(this.getClass().getName()).log(Level.SEVERE, "Unable to serialize Subgroup Request.");
            return null;
        }
    }

    public static class SubgroupRequest implements IJsonSerializable {
        private final transient String listId;
        private final transient String groupField;
        private final transient String optionsField;

        public SubgroupRequest(SubgroupRequestBuilder builder){
            this.listId = builder.listId;
            this.groupField = builder.groupField;
            this.optionsField = builder.optionsField;
        }

        public String getListId() {
            return listId;
        }

        public String getGroupField() {
            return groupField;
        }


        public String getOptionsField() {
            return optionsField;
        }

        @Override
        public String getBody() {
            return new Gson().toJson(this);
        }
    }

    private static class SubgroupRequestSingle extends SubgroupRequest{
        private final String name;
        public SubgroupRequestSingle(SubgroupRequestBuilder builder) {
            super(builder);
            this.name = builder.name;
        }

        public String getName() {
            return name;
        }
    }

    private static class SubgroupRequestMultiple extends SubgroupRequest{
        private final List<String> name;
        public SubgroupRequestMultiple(SubgroupRequestBuilder builder) {
            super(builder);
            this.name = builder.names;
        }

        public List<String> getName() {
            return name;
        }
    }
}
