/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package com.apifon.sdk.resource;

import com.apifon.sdk.Mookee;
import com.apifon.sdk.exception.OAuthTokenException;
import com.apifon.sdk.model.request.OAuthConfigurationBuilder.OAuthConfiguration;
import com.apifon.sdk.resource.helpers.OAuthCredentials;
import com.apifon.sdk.resource.helpers.OAuthGrantType;
import com.apifon.sdk.response.OauthTokenErrorResponse;
import com.google.gson.Gson;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.charset.StandardCharsets;
import java.util.Optional;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.util.EntityUtils;

/**
 *
 * @author Apifon
 */
public class OAuthResource extends AbstractResource {

    private static final String AUTHORIZE_URL = "oauth2/authorize";
    private static final String TOKEN_URL = "oauth2/token";
    private static final String REVOKE_URL = "oauth2/revoke";
    private final OAuthConfiguration authorizationConfig;
    private final OAuthCredentials oauthCredentials;

    public OAuthResource(OAuthConfiguration authorizationConfig) {
        super(Mookee.getInstance().getIdentityServerUrl());
        this.authorizationConfig = authorizationConfig;
        this.oauthCredentials = new OAuthCredentials();
    }

    /**
     * Generates access token for the given client credentials.
     *
     * @throws OAuthTokenException If token generation fails
     */
    public void getTokensForClientCredentialsGrant() throws OAuthTokenException {
        if (!this.getGrantType().equals(OAuthGrantType.CLIENT_CREDENTIALS)) {
            throw new OAuthTokenException("Tried to retrieve tokens for client credentials but grant type is not client_credentials");
        }
        try (CloseableHttpClient client = HttpClientBuilder.create().build()) {
            String endpoint = super.getEndpoint() + TOKEN_URL
                    + "?grant_type=" + this.authorizationConfig.getGrantType().getGrantTypeAsString()
                    + "&client_id=" + this.authorizationConfig.getClientId()
                    + "&client_secret=" + this.authorizationConfig.getClientSecret()
                    + "&scope=" + this.authorizationConfig.getScopesAsString();
            HttpPost post = new HttpPost(new URI(endpoint));
            post.setHeader("Content-Type", "application/x-www-form-urlencoded");
            CloseableHttpResponse response = client.execute(post);
            String respString = EntityUtils.toString(response.getEntity(), StandardCharsets.UTF_8);
            EntityUtils.consume(response.getEntity());
            if (response.getStatusLine().getStatusCode() == 200) {
                OAuthCredentials credentials = new Gson().fromJson(respString, OAuthCredentials.class);
                oauthCredentials.setAccessCode(null);
                oauthCredentials.setAccessToken(credentials.getAccessToken());
                oauthCredentials.setRefreshToken(null);
                oauthCredentials.setExpiresIn(credentials.getExpiresIn());
            } else {
                OauthTokenErrorResponse resp = new Gson().fromJson(respString,OauthTokenErrorResponse.class);
                throw new OAuthTokenException(respString,resp);
            }
        } catch (IOException ex) {
            throw new OAuthTokenException(ex);
        } catch (URISyntaxException ex) {
            throw new OAuthTokenException(new IOException(ex));
        }
    }

    /**
     * Tries to revoke tokens (access token and refresh token if available) for
     * the given resource.<br><br>
     * Note that after this method completes successfully your tokens will be
     * invalid and any authorization must be performed again.
     *
     * @throws OAuthTokenException in case an exception occurs during token revocation.
     */
    public void revokeTokens() throws OAuthTokenException {
        revokeAccessToken();
    }

    /**
     * Revokes any active access token (if any) This method has no effect if
     * there is no access token available.<br><br>
     * <b>NOTE</b> Revoking an active access token revokes the currently associated
     * refresh tokens as well if any.
     *
     * @throws OAuthTokenException In case of a token revocation failure
     * @returns true if the access token is revoked, false 
     */
    private boolean revokeAccessToken() throws OAuthTokenException {
        if (this.oauthCredentials.getAccessToken() != null) {
            revokeToken(this.oauthCredentials.getAccessToken(), "access_token");
            this.oauthCredentials.setAccessToken(null);
            return true;
        }
        return false;
    }

    /**
     * Issue new access tokens<br>
     * <b>NOTE:</b> Old access tokens (if any) are revoked using this
     * method for a given scope with client_credentials.
     *
     * @see #revokeAccessToken() 
     * @throws OAuthTokenException
     */
    public void refreshToken() throws OAuthTokenException {
        if (this.getGrantType() == null) {
            return;
        }
        switch (this.getGrantType()) {
            case CLIENT_CREDENTIALS:
                getTokensForClientCredentialsGrant();
                break;
        }
    }

    private OAuthGrantType getGrantType() {
        return this.authorizationConfig.getGrantType();
    }

    /**
     * <b>Warning: Access tokens are extremely sensitive information granting
     * access to account information, be extremely careful with
     * usages.</b><br><br>
     * <b>Note:</b> Token MAY have expired.
     *
     * @return the current access token or an empty Optional if there is no
     * access token available.
     */
    public Optional<String> getAccessToken() {
        return Optional.ofNullable(this.oauthCredentials.getAccessToken());
    }

    /**
     * <b>Warning: Refresh tokens are extremely sensitive information granting
     * access to account information, be extremely careful with
     * usages.</b><br><br>
     * <b>Note:</b> Token MAY have expired.<br>
     *
     * @return the current refresh token or an empty Optional if there is no
     * refresh token available.
     */
    public Optional<String> getRefreshToken() {
        return Optional.ofNullable(this.oauthCredentials.getRefreshToken());
    }

    private void revokeToken(String token, String tokenTypeHint) throws OAuthTokenException {
        try (CloseableHttpClient client = HttpClientBuilder.create().build()) {
            String endpoint = super.getEndpoint() + REVOKE_URL
                    + "?client_id=" + this.authorizationConfig.getClientId()
                    + "&token=" + token
                    + "&token_type_hint=" + tokenTypeHint;
            if (OAuthGrantType.CLIENT_CREDENTIALS.equals(this.authorizationConfig.getGrantType())) {
                endpoint += "&client_secret=" + this.authorizationConfig.getClientSecret();
            }
            HttpPost post = new HttpPost(new URI(endpoint));
            post.setHeader("Content-Type", "application/x-www-form-urlencoded");
            CloseableHttpResponse response = client.execute(post);
            if (response.getStatusLine().getStatusCode() != 200) {
                String entity = EntityUtils.toString(response.getEntity());
                throw new OAuthTokenException("Failed to revoke token with status " + String.valueOf(response.getStatusLine().getStatusCode()
                        + " and response " + entity));
            }
        } catch (IOException ex) {
            throw new OAuthTokenException("Failed to revoke refresh_token with exception", ex);
        } catch (URISyntaxException ex) {
            throw new OAuthTokenException(new IOException(ex));
        }
    }

    public OAuthCredentials getOauthCredentials() {
        return oauthCredentials;
    }

}
