package com.apifon.sdk.resource;

import com.apifon.sdk.exception.ApifonRestException;
import com.apifon.sdk.model.CollectionEntityResponse;
import com.apifon.sdk.model.request.GroupRequestBuilder;
import com.apifon.sdk.model.request.GroupSingleQuery;
import com.apifon.sdk.model.request.SubgroupRequestBuilder;
import com.apifon.sdk.model.request.SubgroupSingleQuery;
import com.apifon.sdk.response.GroupResponse;
import com.apifon.sdk.response.SubgroupResponse;
import com.google.gson.Gson;
import org.apache.http.HttpEntity;
import org.apache.http.HttpHeaders;
import org.apache.http.HttpResponse;
import org.apache.http.util.EntityUtils;

import java.io.IOException;
import java.net.URI;
import java.util.logging.Level;
import java.util.logging.Logger;

public class SubgroupResource extends AbstractResource implements UriLocator<SubgroupResponse>{

    public SubgroupResource() {
        super("/services/api/v1/list/{list_id}/{group_field}");
    }
    /**
     * @param endpoint the curEndpoint to set
     * @param listId the list ID
     */
    public void replacePlaceholderAndSetUrl(String endpoint, String listId, String groupField) {
        endpoint = endpoint.replace("{list_id}", listId);
        super.setCurEndpoint(endpoint.replace("{group_field}", groupField));
    }
    /**
     * Creates a SubgroupRequest using the Apifon REST API.
     * @param request
     * @return a URI describing the resource created.
     * @throws ApifonRestException When an API exception is encountered
     */
    public URI addSubgroup(SubgroupRequestBuilder.SubgroupRequest request) throws ApifonRestException {
        if (request.getListId() == null || request.getListId().isEmpty()) {
            throw new IllegalArgumentException("request.id is not set.");
        }
        if (request.getGroupField() == null || request.getGroupField().isEmpty()) {
            throw new IllegalArgumentException("request.field is not set.");
        }
        this.setCurMethod("POST");
        this.replacePlaceholderAndSetUrl(this.getEndpoint() + "/option", request.getListId(),request.getGroupField());
        this.setCurBody(request.getBody());
        HttpResponse response = this.dispatchToResponse();
        return URI.create(response.getFirstHeader(HttpHeaders.LOCATION).getValue());
    }

    /**
     * Creates a SubgroupRequest for multiple subgroups using the Apifon REST API.
     * @param request
     * @return a URI describing the resource created.
     * @throws ApifonRestException When an API exception is encountered
     */
    public URI addSubgroups(SubgroupRequestBuilder.SubgroupRequest request) throws ApifonRestException {
        if (request.getListId() == null || request.getListId().isEmpty()) {
            throw new IllegalArgumentException("request.id is not set.");
        }
        if (request.getGroupField() == null || request.getGroupField().isEmpty()) {
            throw new IllegalArgumentException("request.field is not set.");
        }
        this.setCurMethod("POST");
        this.replacePlaceholderAndSetUrl(this.getEndpoint() + "/options", request.getListId(),request.getGroupField());
        this.setCurBody(request.getBody());
        HttpResponse response = this.dispatchToResponse();
        return URI.create(response.getFirstHeader(HttpHeaders.LOCATION).getValue());
    }

    /**
     * Updates an existing SubgroupResponse according to request.
     * @param request
     * @return a URI describing the resource updated.
     * @throws ApifonRestException When an API exception is encountered
     * @throws IllegalArgumentException if the request doesn't contain id.
     */
    public URI updateSubgroup(SubgroupRequestBuilder.SubgroupRequest request) throws ApifonRestException, IllegalArgumentException {
        if (request.getListId() == null || request.getListId().isEmpty()) {
            throw new IllegalArgumentException("request.id is not set.");
        }
        if (request.getGroupField() == null || request.getGroupField().isEmpty()) {
            throw new IllegalArgumentException("request.field is not set.");
        }
        if (request.getOptionsField() == null || request.getOptionsField().isEmpty()) {
            throw new IllegalArgumentException("request.optionsField is not set.");
        }
        this.setCurMethod("PUT");
        this.replacePlaceholderAndSetUrl(this.getEndpoint() + "/" + request.getOptionsField(),request.getListId(),request.getGroupField());
        this.setCurBody(request.getBody());
        HttpResponse response = this.dispatchToResponse();
        return URI.create(response.getFirstHeader(HttpHeaders.LOCATION).getValue());
    }

    /**
     * Used to get a single SubgroupSingleQuery from the API.
     * @param request
     * @return
     * @throws ApifonRestException When an API exception is encountered
     */
    public SubgroupResponse getSubgroup(SubgroupSingleQuery request) throws ApifonRestException {
        if (request.getListId() == null || request.getListId().isEmpty()) {
            throw new IllegalArgumentException("request.id is not set.");
        }
        if (request.getGroupField() == null || request.getGroupField().isEmpty()) {
            throw new IllegalArgumentException("request.field is not set.");
        }
        if (request.getPathId() == null || request.getPathId().isEmpty()) {
            throw new IllegalArgumentException("request.optionsField is not set.");
        }
        Gson gson = new Gson();
        this.setCurMethod("GET");
        this.replacePlaceholderAndSetUrl(this.getEndpoint() + "/option/" + request.getPathId(),request.getListId(),request.getGroupField());
        this.setCurBody("");
        HttpResponse response = this.dispatchToResponse();
        HttpEntity httpEntity = response.getEntity();
        try {
            String responseString = EntityUtils.toString(httpEntity);
            return gson.fromJson(responseString, SubgroupResponse.class);
        } catch (IOException ex) {
            Logger.getLogger(this.getClass().getName()).log(Level.SEVERE, "Exception while trying to parse response", ex);
            throw new ApifonRestException(ex);
        }
    }

    /**
     * Removes the subgroup with the given id <br><br>
     * <b>NOTE: This is a destructive irreversable operation</b>
     *
     * @param listId the list id.
     * @param groupField the group to remove.
     * @return true if removed, false otherwise
     * @throws ApifonRestException When an API exception is encountered
     */
    public boolean removeSubgroup(String listId, String groupField, String optionField) throws ApifonRestException {
        this.setCurMethod("DELETE");
        this.replacePlaceholderAndSetUrl(this.getEndpoint() + "/" + optionField,listId,groupField);
        this.setCurBody("");
        HttpResponse response = this.dispatchToResponse();
        return this.isSuccessful(response);
    }

    @Override
    public SubgroupResponse getByResourceUri(URI resourceUri) throws ApifonRestException {
        Gson gson = new Gson();
        this.setCurMethod("GET");
        this.setCurEndpoint(resourceUri.toASCIIString());
        this.setCurBody("");
        HttpResponse response = this.dispatchToResponse();
        HttpEntity httpEntity = response.getEntity();
        try {
            String responseString = EntityUtils.toString(httpEntity);
            return gson.fromJson(responseString, SubgroupResponse.class);
        } catch (IOException ex) {
            Logger.getLogger(this.getClass().getName()).log(Level.SEVERE, "Exception while trying to parse response", ex);
            throw new ApifonRestException(ex);
        }
    }

    @Override
    public CollectionEntityResponse<SubgroupResponse> getListByResouceUri(URI resourceUri) throws ApifonRestException {
        return null;
    }
}
